@@ -13,11 +13,14 @@
 #include <osmocom/core/stat_item.h>
 #include <osmocom/gsm/gsm48.h>
 #include <osmocom/crypt/auth.h>
+#include <osmocom/crypt/utran_cipher.h>
 
 #include <osmocom/mgcp_client/mgcp_client.h>
+#include <osmocom/mgcp_client/mgcp_client_pool.h>
 
 #include <osmocom/msc/msc_common.h>
 #include <osmocom/msc/neighbor_ident.h>
+#include <osmocom/msc/sms_queue.h>
 
 #include "gsm_data_shared.h"
 #include "osmux.h"
@@ -30,6 +33,7 @@ struct vlr_instance;
 struct vlr_subscr;
 struct gsup_client_mux;
 
+#define SMS_DEFAULT_DB_FILE_PATH "sms.db"
 #define tmsi_from_string(str) strtoul(str, NULL, 10)
 
 enum {
@@ -43,6 +47,8 @@ enum {
 	MSC_CTR_CM_SERVICE_REQUEST_ACCEPTED,
 	MSC_CTR_PAGING_RESP_REJECTED,
 	MSC_CTR_PAGING_RESP_ACCEPTED,
+	MSC_CTR_CM_RE_ESTABLISH_REQ_REJECTED,
+	MSC_CTR_CM_RE_ESTABLISH_REQ_ACCEPTED,
 	MSC_CTR_SMS_SUBMITTED,
 	MSC_CTR_SMS_NO_RECEIVER,
 	MSC_CTR_SMS_DELIVERED,
@@ -75,6 +81,8 @@ static const struct rate_ctr_desc msc_ctr_description[] = {
 	[MSC_CTR_CM_SERVICE_REQUEST_ACCEPTED] = {"cm_service_request:accepted", "Accepted CM Service Requests."},
 	[MSC_CTR_PAGING_RESP_REJECTED] = 	{"paging_resp:rejected", "Rejected Paging Responses."},
 	[MSC_CTR_PAGING_RESP_ACCEPTED] = 	{"paging_resp:accepted", "Accepted Paging Responses."},
+	[MSC_CTR_CM_RE_ESTABLISH_REQ_REJECTED] = {"cm_re_establish_request:rejected", "Rejected CM Re-Establishing Requests."},
+	[MSC_CTR_CM_RE_ESTABLISH_REQ_ACCEPTED] = {"cm_re_establish_request:accepted", "Accepted CM Re-Establishing Requests."},
 	[MSC_CTR_SMS_SUBMITTED] = 		{"sms:submitted", "Total MO SMS received from the MS."},
 	[MSC_CTR_SMS_NO_RECEIVER] = 		{"sms:no_receiver", "Failed MO SMS delivery attempts (no receiver found)."},
 	[MSC_CTR_SMS_DELIVER_UNKNOWN_ERROR] =	{"sms:deliver_unknown_error", "Failed MO SMS delivery attempts (other reason)."},
@@ -150,10 +158,8 @@ struct gsm_network {
 	bool authentication_required;
 	int send_mm_info;
 
-	/* Whether to use encryption on UTRAN.
-	 * TODO: we should offer a choice of UEA1 and/or UEA2, and probably replace this bool with a bit-mask of
-	 * permitted Iu encryption algorithms. See also OS#4143 and the 'encryption uea' vty command. */
-	bool uea_encryption;
+	/* bit-mask of permitted encryption algorithms. LSB=UEA0, MSB=UEA7 */
+	uint8_t uea_encryption_mask;
 
 	struct rate_ctr_group *msc_ctrs;
 	struct osmo_stat_item_group *statg;
@@ -171,8 +177,6 @@ struct gsm_network {
 	 */
 	struct llist_head trans_list;
 
-	unsigned int paging_response_timer;
-
 	/* Radio Resource Location Protocol (TS 04.31) */
 	struct {
 		enum rrlp_mode mode;
@@ -205,9 +209,6 @@ struct gsm_network {
 
 	struct vlr_instance *vlr;
 
-	/* Periodic location update default value */
-	uint8_t t3212;
-
 	/* Global MNCC guard timer value */
 	int mncc_guard_timeout;
 	/* Global guard timer value for NCSS sessions */
@@ -215,8 +216,10 @@ struct gsm_network {
 
 	struct {
 		struct osmo_tdef *tdefs;
-		struct mgcp_client_conf conf;
-		struct mgcp_client *client;
+		struct mgcp_client_conf *conf;
+		/* MGW pool, also includes the single MGCP client as fallback if no
+		 * pool is configured. */
+		struct mgcp_client_pool *mgw_pool;
 	} mgw;
 
 	struct {
@@ -234,10 +237,6 @@ struct gsm_network {
 		struct sccp_ran_inst *sri;
 	} a;
 
-	/* A list of neighbor BSCs. This list is defined statically via VTY and does not
-	* necessarily correspond to BSCs attached to the A interface at a given moment. */
-	struct neighbor_ident_list *neighbor_list;
-
 	struct {
 		/* MSISDN to which to route MO emergency calls */
 		char *route_to_msisdn;
@@ -249,6 +248,8 @@ struct gsm_network {
 	 * and will be of the form 'MSC-00-00-00-00-00-00' */
 	char *msc_ipa_name;
 
+	/* A list of neighbor BSCs. This list is defined statically via VTY and does not
+	* necessarily correspond to BSCs attached to the A interface at a given moment. */
 	struct llist_head neighbor_ident_list;
 
 	struct {
@@ -262,9 +263,21 @@ struct gsm_network {
 
 	/* Whether to use call waiting on the network */
 	bool call_waiting;
+
+	/* Whether to use lcls on the network */
+	bool lcls_permitted;
+
+	/* SMS queue config parameters */
+	struct sms_queue_config *sms_queue_cfg;
+
+	/* ASCI feature support */
+	struct {
+		bool enable;
+		struct llist_head gcr_lists;
+	} asci;
 };
 
-struct osmo_esme;
+struct smpp_esme;
 
 enum gsm_sms_source_id {
 	SMS_SOURCE_UNKNOWN = 0,
@@ -293,7 +306,7 @@ struct gsm_sms {
 	} gsm411;
 
 	struct {
-		struct osmo_esme *esme;
+		struct smpp_esme *esme;
 		uint32_t sequence_nr;
 		int transaction_mode;
 		char msg_id[16];
