@@ -43,14 +43,14 @@
 #include <osmocom/msc/gsm_04_80.h>
 #include <osmocom/msc/gsm_subscriber.h>
 #include <osmocom/msc/debug.h>
-#include <osmocom/msc/vlr.h>
+#include <osmocom/vlr/vlr.h>
 #include <osmocom/msc/gsm_04_08.h>
 #include <osmocom/msc/transaction.h>
 #include <osmocom/gsupclient/gsup_client.h>
 #include <osmocom/msc/msc_a.h>
 #include <osmocom/msc/msub.h>
 #include <osmocom/msc/paging.h>
-#include <osmocom/msc/gsup_client_mux.h>
+#include <osmocom/gsupclient/gsup_client_mux.h>
 
 /* FIXME: choose a proper range */
 static uint32_t new_callref = 0x20000001;
@@ -142,6 +142,9 @@ int gsm0911_rcv_nc_ss(struct msc_a *msc_a, struct msgb *msg)
 			gsm48_tx_simple(msc_a,
 				GSM48_PDISC_NC_SS | (tid << 4),
 				GSM0480_MTYPE_RELEASE_COMPLETE);
+			/* Decrement use counter that has been incremented by CM Service Request (SS).
+			 * If there is no other service request, the BSS connection will be released. */
+			msc_a_put(msc_a, MSC_A_USE_CM_SERVICE_SS);
 			return -EINVAL;
 		}
 
@@ -151,6 +154,9 @@ int gsm0911_rcv_nc_ss(struct msc_a *msc_a, struct msgb *msg)
 			gsm48_tx_simple(msc_a,
 				GSM48_PDISC_NC_SS | (tid << 4),
 				GSM0480_MTYPE_RELEASE_COMPLETE);
+			/* Decrement use counter that has been incremented by CM Service Request (SS).
+			 * If there is no other service request, the BSS connection will be released. */
+			msc_a_put(msc_a, MSC_A_USE_CM_SERVICE_SS);
 			return -ENOMEM;
 		}
 
@@ -441,7 +447,7 @@ int gsm0911_gsup_rx(struct gsup_client_mux *gcm, void *data, const struct osmo_g
 	log_set_context(LOG_CTX_VLR_SUBSCR, vsub);
 
 	/* Attempt to find DTAP-transaction */
-	trans = trans_find_by_callref(net, gsup_msg->session_id);
+	trans = trans_find_by_callref(net, TRANS_USSD, gsup_msg->session_id);
 
 	/* Handle errors */
 	if (OSMO_GSUP_IS_MSGT_ERROR(gsup_msg->message_type)) {
