@@ -93,7 +93,7 @@
 #include <osmocom/msc/msub.h>
 #include <osmocom/msc/debug.h>
 #include <osmocom/msc/msc_a.h>
-#include <osmocom/msc/vlr.h>
+#include <osmocom/vlr/vlr.h>
 #include <osmocom/msc/rtp_stream.h>
 #include <osmocom/msc/codec_mapping.h>
 #include <osmocom/msc/msc_vgcs.h>
