@@ -1,5 +1,5 @@
 /*
- * (C) 2019 by sysmocom - s.m.f.c. GmbH <info@sysmocom.de>
+ * (C) 2019 by sysmocom - s.f.m.c. GmbH <info@sysmocom.de>
  * All Rights Reserved
  *
  * SPDX-License-Identifier: AGPL-3.0+
@@ -30,9 +30,10 @@
 #include <osmocom/msc/msub.h>
 #include <osmocom/msc/msc_i.h>
 #include <osmocom/msc/msc_a.h>
-#include <osmocom/msc/vlr.h>
+#include <osmocom/vlr/vlr.h>
 #include <osmocom/msc/ran_conn.h>
 #include <osmocom/msc/cell_id_list.h>
+#include <osmocom/msc/msc_vgcs.h>
 
 static struct osmo_fsm ran_peer_fsm;
 
@@ -80,13 +81,13 @@ static struct ran_peer *ran_peer_alloc(struct sccp_ran_inst *sri, const struct o
 
 struct ran_peer *ran_peer_find_or_create(struct sccp_ran_inst *sri, const struct osmo_sccp_addr *peer_addr)
 {
-	struct ran_peer *rp = ran_peer_find(sri, peer_addr);
+	struct ran_peer *rp = ran_peer_find_by_addr(sri, peer_addr);
 	if (rp)
 		return rp;
 	return ran_peer_alloc(sri, peer_addr);
 }
 
-struct ran_peer *ran_peer_find(struct sccp_ran_inst *sri, const struct osmo_sccp_addr *peer_addr)
+struct ran_peer *ran_peer_find_by_addr(struct sccp_ran_inst *sri, const struct osmo_sccp_addr *peer_addr)
 {
 	struct ran_peer *rp;
 	llist_for_each_entry(rp, &sri->ran_peers, entry) {
@@ -118,29 +119,28 @@ void ran_peer_discard_all_conns(struct ran_peer *rp)
 	struct ran_conn *conn, *next;
 
 	ran_peer_for_each_ran_conn_safe(conn, next, rp) {
-		ran_conn_discard(conn);
+		/* Tell VGCS FSM that the connections have been cleared. */
+		if (conn->vgcs.bss)
+			vgcs_vbs_clear_cpl(conn->vgcs.bss, NULL);
+		else if (conn->vgcs.cell)
+			vgcs_vbs_clear_cpl_channel(conn->vgcs.cell, NULL);
+		else ran_conn_discard(conn);
 	}
 }
 
-/* TODO: create an sccp_ran_ops.rx_reset(_ack) to handle this differently on 2g and 3G */
-/* We expect RAN peer to provide use with an Osmocom extension TLV in BSSMAP_RESET to
- * announce Osmux support */
-static void ran_peer_update_osmux_support(struct ran_peer *rp, struct msgb *msg)
+static void ran_peer_update_osmux_support(struct ran_peer *rp, int supports_osmux)
 {
-	struct tlv_parsed tp;
-	int rc;
 	bool old_value = rp->remote_supports_osmux;
 
-	OSMO_ASSERT(msg);
-	msg->l3h = msg->l2h + sizeof(struct bssmap_header);
-	rc = tlv_parse(&tp, gsm0808_att_tlvdef(), msg->l3h + 1, msgb_l3len(msg) - 1, 0, 0);
-	if (rc < 0)
-		LOG_RAN_PEER(rp, LOGL_NOTICE, "Failed parsing TLV looking for Osmux support\n");
-
-	if (TLVP_PRESENT(&tp, GSM0808_IE_OSMO_OSMUX_SUPPORT)) {
+	switch (supports_osmux) {
+	case 1:
 		rp->remote_supports_osmux = true;
-	} else {
+		break;
+	case -1:
 		rp->remote_supports_osmux = false;
+		break;
+	default:
+		return;
 	}
 
 	if (old_value != rp->remote_supports_osmux)
@@ -155,8 +155,6 @@ static void ran_peer_rx_reset(struct ran_peer *rp, struct msgb* msg)
 
 	ran_peer_discard_all_conns(rp);
 
-	ran_peer_update_osmux_support(rp, msg);
-
 	reset_ack = rp->sri->ran->sccp_ran_ops.make_reset_msg(rp->sri, SCCP_RAN_MSG_RESET_ACK);
 
 	if (!reset_ack) {
@@ -183,7 +181,6 @@ static void ran_peer_rx_reset(struct ran_peer *rp, struct msgb* msg)
 static void ran_peer_rx_reset_ack(struct ran_peer *rp, struct msgb* msg)
 {
 	ran_peer_state_chg(rp, RAN_PEER_ST_READY);
-	ran_peer_update_osmux_support(rp, msg);
 }
 
 void ran_peer_reset(struct ran_peer *rp)
@@ -213,14 +210,18 @@ void ran_peer_allstate_action(struct osmo_fsm_inst *fi, uint32_t event, void *da
 	struct ran_peer *rp = fi->priv;
 	struct ran_peer_ev_ctx *ctx = data;
 	struct msgb *msg = ctx->msg;
+	enum reset_msg_type is_reset;
+	int supports_osmux;
 
 	switch (event) {
 	case RAN_PEER_EV_MSG_UP_CL:
-		switch (rp->sri->ran->sccp_ran_ops.is_reset_msg(rp->sri, msg)) {
-		case 1:
+		is_reset = rp->sri->ran->sccp_ran_ops.is_reset_msg(rp->sri, fi, msg, &supports_osmux);
+		ran_peer_update_osmux_support(rp, supports_osmux);
+		switch (is_reset) {
+		case SCCP_RAN_MSG_RESET:
 			osmo_fsm_inst_dispatch(fi, RAN_PEER_EV_RX_RESET, msg);
 			return;
-		case 2:
+		case SCCP_RAN_MSG_RESET_ACK:
 			osmo_fsm_inst_dispatch(fi, RAN_PEER_EV_RX_RESET_ACK, msg);
 			return;
 		default:
@@ -398,18 +399,22 @@ void ran_peer_st_ready(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 		OSMO_ASSERT(ctx->conn);
 		OSMO_ASSERT(ctx->msg);
 
-		if (!ctx->conn->msc_role) {
+		if (ctx->conn->msc_role) {
+			/* "normal" A connection, dispatch to MSC-I or MSC-T */
+			an_apdu = (struct an_apdu){
+				.an_proto = rp->sri->ran->an_proto,
+				.msg = ctx->msg,
+			};
+			osmo_fsm_inst_dispatch(ctx->conn->msc_role, MSC_EV_FROM_RAN_UP_L2, &an_apdu);
+		} else if (ctx->conn->vgcs.bss) {
+			/* VGCS call related */
+			msc_a_rx_vgcs_bss(ctx->conn->vgcs.bss, ctx->conn, ctx->msg);
+		} else if (ctx->conn->vgcs.cell) {
+			/* VGCS channel related */
+			msc_a_rx_vgcs_cell(ctx->conn->vgcs.cell, ctx->conn, ctx->msg);
+		} else
 			LOG_RAN_PEER(rp, LOGL_ERROR,
 				     "Rx CO message on conn that is not associated with any MSC role\n");
-			return;
-		}
-
-		an_apdu = (struct an_apdu){
-			.an_proto = rp->sri->ran->an_proto,
-			.msg = ctx->msg,
-		};
-
-		osmo_fsm_inst_dispatch(ctx->conn->msc_role, MSC_EV_FROM_RAN_UP_L2, &an_apdu);
 		return;
 
 	case RAN_PEER_EV_MSG_DOWN_CO_INITIAL:
@@ -629,17 +634,6 @@ struct ran_peer *ran_peer_find_by_cell_id(struct sccp_ran_inst *sri, const struc
 	return found;
 }
 
-struct ran_peer *ran_peer_find_by_addr(struct sccp_ran_inst *sri, const struct osmo_sccp_addr *addr)
-{
-	struct ran_peer *rp;
-
-	llist_for_each_entry(rp, &sri->ran_peers, entry) {
-		if (!osmo_sccp_addr_ri_cmp(addr, &rp->peer_addr))
-			return rp;
-	}
-	return NULL;
-}
-
 int ran_peers_down_paging(struct sccp_ran_inst *sri, enum CELL_IDENT page_where, struct vlr_subscr *vsub,
 			  enum paging_cause cause)
 {
