@@ -24,11 +24,12 @@
 #include <osmocom/msc/debug.h>
 #include <osmocom/core/talloc.h>
 #include <osmocom/msc/gsm_04_08.h>
-#include <osmocom/msc/vlr.h>
+#include <osmocom/vlr/vlr.h>
 #include <osmocom/msc/msc_a.h>
 #include <osmocom/msc/msub.h>
 #include <osmocom/msc/paging.h>
 #include <osmocom/msc/silent_call.h>
+#include <osmocom/msc/msc_vgcs.h>
 
 void *tall_trans_ctx;
 
@@ -73,16 +74,17 @@ struct gsm_trans *trans_find_by_id(const struct msc_a *msc_a,
 
 /*! Find a transaction by call reference
  * \param[in] net Network in which we should search
+ * \param[in] type Transaction type (e.g. TRANS_CC)
  * \param[in] callref Call Reference of transaction
  * \returns Matching transaction, if any
  */
-struct gsm_trans *trans_find_by_callref(const struct gsm_network *net,
+struct gsm_trans *trans_find_by_callref(const struct gsm_network *net, enum trans_type type,
 					uint32_t callref)
 {
 	struct gsm_trans *trans;
 
 	llist_for_each_entry(trans, &net->trans_list, entry) {
-		if (trans->callref == callref)
+		if (trans->callref == callref && trans->type == type)
 			return trans;
 	}
 	return NULL;
@@ -122,7 +124,8 @@ struct osmo_lcls *trans_lcls_compose(const struct gsm_trans *trans, bool use_lac
 
 	struct osmo_ss7_instance *ss7 = osmo_sccp_get_ss7(trans->net->a.sri->sccp);
 	struct osmo_lcls *lcls;
-	uint8_t w = osmo_ss7_pc_width(&ss7->cfg.pc_fmt);
+	const struct osmo_ss7_pc_fmt *pc_fmt = osmo_ss7_instance_get_pc_fmt(ss7);
+	uint8_t w = osmo_ss7_pc_width(pc_fmt);
 
 	if (!trans->net->lcls_permitted) {
 		LOGP(DCC, LOGL_NOTICE, "LCLS disabled globally\n");
@@ -148,16 +151,22 @@ struct osmo_lcls *trans_lcls_compose(const struct gsm_trans *trans, bool use_lac
 	LOGP(DCC, LOGL_INFO, "LCLS: using %u bits (%u bytes) for node ID\n", w, w / 8);
 
 	lcls->gcr.net_len = 3;
-	lcls->gcr.node = ss7->cfg.primary_pc;
+	lcls->gcr.node = osmo_ss7_instance_get_primary_pc(ss7);
 
 	/* net id from Q.1902.3 3-5 bytes, this function gives 3 bytes exactly */
-	osmo_plmn_to_bcd(lcls->gcr.net, &trans->net->plmn);
+	osmo_plmn_to_bcd(lcls->gcr.net, &trans->msc_a->via_cell.lai.plmn);
 
-	osmo_store32be(trans->callref, lcls->gcr.cr);
-	osmo_store16be(use_lac ? trans->msc_a->via_cell.lai.lac : trans->msc_a->via_cell.cell_identity, lcls->gcr.cr + 3);
 
-	LOGP(DCC, LOGL_INFO, "LCLS: allocated %s-based CR-ID %s\n", use_lac ? "LAC" : "CI",
-	     osmo_hexdump(lcls->gcr.cr, 5));
+	/* TS 29.205 Table B.2.1.9.2 Call Reference ID
+	 * 3 octets Call ID + 2 octets BSS ID
+	 */
+	lcls->gcr.cr[2] = (trans->callref >>  0) & 0xff;
+	lcls->gcr.cr[1] = (trans->callref >>  8) & 0xff;
+	lcls->gcr.cr[0] = (trans->callref >> 16) & 0xff;
+	osmo_store16be(use_lac ? trans->msc_a->via_cell.lai.lac : trans->msc_a->via_cell.cell_identity, &lcls->gcr.cr[3]);
+
+	LOGP(DCC, LOGL_INFO, "LCLS: allocated %s-based CR-ID %sfor callref 0x%04x\n", use_lac ? "LAC" : "CI",
+	     osmo_hexdump(lcls->gcr.cr, 5), trans->callref);
 
 	lcls->config = GSM0808_LCLS_CFG_BOTH_WAY;
 	lcls->control = GSM0808_LCLS_CSC_CONNECT;
@@ -175,8 +184,10 @@ static const char *trans_vsub_use(enum trans_type type)
 	return get_value_string_or_null(trans_type_names, type) ? : "trans-type-unknown";
 }
 
+static uint32_t new_call_id = 1;
+
 /*! Allocate a new transaction and add it to network list
- *  \param[in] net Netwokr in which we allocate transaction
+ *  \param[in] net Network in which we allocate transaction
  *  \param[in] subscr Subscriber for which we allocate transaction
  *  \param[in] protocol Protocol (CC/SMS/...)
  *  \param[in] callref Call Reference
@@ -190,8 +201,8 @@ struct gsm_trans *trans_alloc(struct gsm_network *net,
 	int subsys = trans_log_subsys(type);
 	struct gsm_trans *trans;
 
-	/* a valid subscriber is indispensable */
-	if (vsub == NULL) {
+	/* A valid subscriber is indispensable, except for voice group/broadcast calls. */
+	if (vsub == NULL && type != TRANS_GCC && type != TRANS_BCC) {
 		LOGP(subsys, LOGL_ERROR, "unable to alloc transaction, invalid subscriber (NULL)\n");
 		return NULL;
 	}
@@ -206,13 +217,15 @@ struct gsm_trans *trans_alloc(struct gsm_network *net,
 		.log_subsys = subsys,
 		.transaction_id = trans_id,
 		.callref = callref,
+		.call_id = new_call_id++,
 		.net = net,
 		/* empty bearer_cap: make sure the speech_ver array is empty */
 		.bearer_cap = {
 			.speech_ver = { -1 },
 		},
 	};
-	vlr_subscr_get(vsub, trans_vsub_use(type));
+	if (vsub)
+		vlr_subscr_get(vsub, trans_vsub_use(type));
 	llist_add_tail(&trans->entry, &net->trans_list);
 
 	LOG_TRANS(trans, LOGL_DEBUG, "New transaction\n");
@@ -230,6 +243,14 @@ void trans_free(struct gsm_trans *trans)
 	LOG_TRANS(trans, LOGL_DEBUG, "Freeing transaction\n");
 
 	switch (trans->type) {
+	case TRANS_GCC:
+		gsm44068_bcc_gcc_trans_free(trans);
+		usage_token = MSC_A_USE_GCC;
+		break;
+	case TRANS_BCC:
+		gsm44068_bcc_gcc_trans_free(trans);
+		usage_token = MSC_A_USE_BCC;
+		break;
 	case TRANS_CC:
 		_gsm48_cc_trans_free(trans);
 		usage_token = MSC_A_USE_CC;
@@ -341,6 +362,8 @@ void trans_conn_closed(const struct msc_a *msc_a)
 }
 
 const struct value_string trans_type_names[] = {
+	{ TRANS_GCC, "GCC" },
+	{ TRANS_BCC, "BCC" },
 	{ TRANS_CC, "CC" },
 	{ TRANS_SMS, "SMS" },
 	{ TRANS_USSD, "NCSS" },
@@ -351,6 +374,10 @@ const struct value_string trans_type_names[] = {
 uint8_t trans_type_to_gsm48_proto(enum trans_type type)
 {
 	switch (type) {
+	case TRANS_GCC:
+		return GSM48_PDISC_GROUP_CC;
+	case TRANS_BCC:
+		return GSM48_PDISC_BCAST_CC;
 	case TRANS_CC:
 	case TRANS_SILENT_CALL:
 		return GSM48_PDISC_CC;
@@ -375,6 +402,12 @@ const char *trans_name(const struct gsm_trans *trans)
 			 trans_type_name(trans->type), gsm48_cc_state_name(trans->cc.state));
 		return namebuf;
 
+	case TRANS_GCC:
+	case TRANS_BCC:
+		snprintf(namebuf, sizeof(namebuf), "%s:%s",
+			 trans_type_name(trans->type), gsm44068_group_id_string(trans->callref));
+		return namebuf;
+
 	default:
 		return trans_type_name(trans->type);
 	}
