@@ -20,10 +20,10 @@
  */
 
 #include <osmocom/core/fsm.h>
+#include <osmocom/core/tdef.h>
 #include <osmocom/gsm/gsup.h>
 #include <osmocom/gsm/gsm48.h>
-#include <osmocom/msc/vlr.h>
-#include <osmocom/msc/debug.h>
+#include <osmocom/vlr/vlr.h>
 
 #include "vlr_core.h"
 #include "vlr_auth_fsm.h"
@@ -51,6 +51,18 @@ static const struct value_string proc_arq_vlr_event_names[] = {
 	{ 0, NULL }
 };
 
+struct osmo_tdef_state_timeout msc_parq_tdef_states[32] = {
+	[PR_ARQ_S_WAIT_CHECK_IMEI]	= { .T = 3270 },
+	[PR_ARQ_S_WAIT_OBTAIN_IMSI]	= { .T = 3270 },
+};
+
+struct osmo_tdef_state_timeout sgsn_parq_tdef_states[32] = {
+	[PR_ARQ_S_WAIT_CHECK_IMEI]	= { .T = 3370 },
+	[PR_ARQ_S_WAIT_OBTAIN_IMSI]	= { .T = 3370 },
+};
+
+struct osmo_tdef_state_timeout *parq_fsm_state_tdef;
+
 struct proc_arq_priv {
 	struct vlr_instance *vlr;
 	struct vlr_subscr *vsub;
@@ -202,8 +214,7 @@ static void _proc_arq_vlr_post_trace(struct osmo_fsm_inst *fi)
 	if (0 /* IMEI check required */) {
 		/* Chck_IMEI_VLR */
 		vlr->ops.tx_id_req(par->msc_conn_ref, GSM_MI_TYPE_IMEI);
-		osmo_fsm_inst_state_chg(fi, PR_ARQ_S_WAIT_CHECK_IMEI,
-					vlr_timer(vlr, 3270), 3270);
+		osmo_tdef_fsm_inst_state_chg(fi, PR_ARQ_S_WAIT_CHECK_IMEI, parq_fsm_state_tdef, vlr_tdefs, -1);
 	} else
 		_proc_arq_vlr_post_imei(fi);
 }
@@ -405,8 +416,7 @@ static void proc_arq_vlr_fn_init(struct osmo_fsm_inst *fi,
 		/* TMSI was included, are we permitted to use it? */
 		if (vlr->cfg.parq_retrieve_imsi) {
 			/* Obtain_IMSI_VLR */
-			osmo_fsm_inst_state_chg(fi, PR_ARQ_S_WAIT_OBTAIN_IMSI,
-						vlr_timer(vlr, 3270), 3270);
+			osmo_tdef_fsm_inst_state_chg(fi, PR_ARQ_S_WAIT_OBTAIN_IMSI, parq_fsm_state_tdef, vlr_tdefs, -1);
 			return;
 		} else {
 			/* Set User Error: Unidentified Subscriber */
@@ -650,7 +660,7 @@ static struct osmo_fsm proc_arq_vlr_fsm = {
 	.num_states = ARRAY_SIZE(proc_arq_vlr_states),
 	.allstate_event_mask = 0,
 	.allstate_action = NULL,
-	.log_subsys = DVLR,
+	.log_subsys = DLGLOBAL,
 	.event_names = proc_arq_vlr_event_names,
 	.cleanup = proc_arq_vlr_cleanup,
 };
@@ -804,8 +814,18 @@ static struct osmo_fsm upd_loc_child_vlr_fsm = {
 };
 #endif
 
-void vlr_parq_fsm_init(void)
+void vlr_parq_fsm_init(bool is_ps)
 {
+	if (is_ps)
+		parq_fsm_state_tdef = sgsn_parq_tdef_states;
+	else
+		parq_fsm_state_tdef = msc_parq_tdef_states;
+
 	//OSMO_ASSERT(osmo_fsm_register(&upd_loc_child_vlr_fsm) == 0);
 	OSMO_ASSERT(osmo_fsm_register(&proc_arq_vlr_fsm) == 0);
 }
+
+void vlr_parq_fsm_set_log_subsys(int log_subsys)
+{
+	proc_arq_vlr_fsm.log_subsys = log_subsys;
+}
