@@ -20,9 +20,9 @@
  */
 
 #include <osmocom/core/fsm.h>
+#include <osmocom/core/tdef.h>
 #include <osmocom/gsm/gsm48.h>
-#include <osmocom/msc/vlr.h>
-#include <osmocom/msc/debug.h>
+#include <osmocom/vlr/vlr.h>
 
 #include "vlr_core.h"
 #include "vlr_auth_fsm.h"
@@ -150,7 +150,7 @@ static struct osmo_fsm upd_hlr_vlr_fsm = {
 	.num_states = ARRAY_SIZE(upd_hlr_vlr_states),
 	.allstate_event_mask = 0,
 	.allstate_action = NULL,
-	.log_subsys = DVLR,
+	.log_subsys = DLGLOBAL,
 	.event_names = upd_hlr_vlr_event_names,
 };
 
@@ -262,7 +262,7 @@ static struct osmo_fsm sub_pres_vlr_fsm = {
 	.num_states = ARRAY_SIZE(sub_pres_vlr_states),
 	.allstate_event_mask = 0,
 	.allstate_action = NULL,
-	.log_subsys = DVLR,
+	.log_subsys = DLGLOBAL,
 	.event_names = sub_pres_vlr_event_names,
 };
 
@@ -329,6 +329,20 @@ static const struct value_string lu_compl_vlr_event_names[] = {
 	{ 0, NULL }
 };
 
+struct osmo_tdef_state_timeout msc_lu_compl_tdef_states[32] = {
+	[LU_COMPL_VLR_S_WAIT_IMEI]	= { .T = 3270 },
+	[LU_COMPL_VLR_S_WAIT_IMEI_TMSI]	= { .T = 3270 },
+	[LU_COMPL_VLR_S_WAIT_TMSI_CNF]	= { .T = 3250 },
+};
+
+struct osmo_tdef_state_timeout sgsn_lu_compl_tdef_states[32] = {
+	[LU_COMPL_VLR_S_WAIT_IMEI]	= { .T = 3370 },
+	[LU_COMPL_VLR_S_WAIT_IMEI_TMSI]	= { .T = 3370 },
+	[LU_COMPL_VLR_S_WAIT_TMSI_CNF]	= { .T = 3350 },
+};
+
+struct osmo_tdef_state_timeout *lu_compl_fsm_state_tdef;
+
 struct lu_compl_vlr_priv {
 	struct vlr_subscr *vsub;
 	void *msc_conn_ref;
@@ -339,6 +353,7 @@ struct lu_compl_vlr_priv {
 	enum vlr_fsm_result result;
 	uint8_t cause;
 	bool assign_tmsi;
+	enum vlr_lu_type lu_type;
 };
 
 static inline struct lu_compl_vlr_priv *lu_compl_vlr_fi_priv(struct osmo_fsm_inst *fi);
@@ -431,10 +446,9 @@ static void lu_compl_vlr_new_tmsi(struct osmo_fsm_inst *fi)
 		return;
 	}
 
-	osmo_fsm_inst_state_chg(fi, LU_COMPL_VLR_S_WAIT_TMSI_CNF,
-				vlr_timer(vlr, 3250), 3250);
+	osmo_tdef_fsm_inst_state_chg(fi, LU_COMPL_VLR_S_WAIT_TMSI_CNF, lu_compl_fsm_state_tdef, vlr_tdefs, -1);
 
-	vlr->ops.tx_lu_acc(lcvp->msc_conn_ref, vsub->tmsi_new);
+	vlr->ops.tx_lu_acc(lcvp->msc_conn_ref, vsub->tmsi_new, lcvp->lu_type);
 }
 
 /* After completion of Subscriber_Present_VLR */
@@ -455,11 +469,11 @@ static void lu_compl_vlr_wait_subscr_pres(struct osmo_fsm_inst *fi,
 	/* If imeisv_early is enabled: IMEI already retrieved and checked (vlr_loc_upd_node1_pre), don't do it again. */
 	if (vlr->cfg.check_imei_rqd && !vlr->cfg.retrieve_imeisv_early) {
 		/* Check IMEI VLR */
-		osmo_fsm_inst_state_chg(fi,
+		osmo_tdef_fsm_inst_state_chg(fi,
 					lcvp->assign_tmsi ?
 					  LU_COMPL_VLR_S_WAIT_IMEI_TMSI
 					: LU_COMPL_VLR_S_WAIT_IMEI,
-					vlr_timer(vlr, 3270), 3270);
+					lu_compl_fsm_state_tdef, vlr_tdefs, -1);
 		vlr->ops.tx_id_req(lcvp->msc_conn_ref, GSM_MI_TYPE_IMEI);
 		return;
 	}
@@ -473,7 +487,7 @@ static void lu_compl_vlr_wait_subscr_pres(struct osmo_fsm_inst *fi,
 	vsub->tmsi = GSM_RESERVED_TMSI;
 
 	/* Location Updating Accept */
-	vlr->ops.tx_lu_acc(lcvp->msc_conn_ref, GSM_RESERVED_TMSI);
+	vlr->ops.tx_lu_acc(lcvp->msc_conn_ref, GSM_RESERVED_TMSI, lcvp->lu_type);
 	vlr_lu_compl_fsm_success(fi);
 }
 
@@ -520,7 +534,7 @@ static void lu_compl_vlr_wait_imei(struct osmo_fsm_inst *fi, uint32_t event,
 	vsub->tmsi = GSM_RESERVED_TMSI;
 
 	/* No TMSI needed, accept now. */
-	vlr->ops.tx_lu_acc(lcvp->msc_conn_ref, GSM_RESERVED_TMSI);
+	vlr->ops.tx_lu_acc(lcvp->msc_conn_ref, GSM_RESERVED_TMSI, lcvp->lu_type);
 	vlr_lu_compl_fsm_success(fi);
 }
 
@@ -600,7 +614,7 @@ static struct osmo_fsm lu_compl_vlr_fsm = {
 	.num_states = ARRAY_SIZE(lu_compl_vlr_states),
 	.allstate_event_mask = 0,
 	.allstate_action = NULL,
-	.log_subsys = DVLR,
+	.log_subsys = DLGLOBAL,
 	.event_names = lu_compl_vlr_event_names,
 };
 
@@ -616,7 +630,8 @@ lu_compl_vlr_proc_alloc(struct osmo_fsm_inst *parent,
 			void *msc_conn_ref,
 			uint32_t parent_event_success,
 			uint32_t parent_event_failure,
-			bool assign_tmsi)
+			bool assign_tmsi,
+			enum vlr_lu_type lu_type)
 {
 	struct osmo_fsm_inst *fi;
 	struct lu_compl_vlr_priv *lcvp;
@@ -632,6 +647,7 @@ lu_compl_vlr_proc_alloc(struct osmo_fsm_inst *parent,
 	lcvp->parent_event_success = parent_event_success;
 	lcvp->parent_event_failure = parent_event_failure;
 	lcvp->assign_tmsi = assign_tmsi;
+	lcvp->lu_type = lu_type;
 	fi->priv = lcvp;
 
 	return fi;
@@ -663,6 +679,20 @@ static const struct value_string fsm_lu_event_names[] = {
 	{ 0, NULL }
 };
 
+struct osmo_tdef_state_timeout msc_lu_tdef_states[32] = {
+	[VLR_ULA_S_WAIT_IMEISV] = { .T = 3270 },
+	[VLR_ULA_S_WAIT_HLR_CHECK_IMEI_EARLY] = { .T = 3270 },
+	[VLR_ULA_S_WAIT_IMSI] = { .T = 3270 },
+};
+
+struct osmo_tdef_state_timeout sgsn_lu_tdef_states[32] = {
+	[VLR_ULA_S_WAIT_IMEISV] = { .T = 3370 },
+	[VLR_ULA_S_WAIT_HLR_CHECK_IMEI_EARLY] = { .T = 3370 },
+	[VLR_ULA_S_WAIT_IMSI] = { .T = 3370 },
+};
+
+struct osmo_tdef_state_timeout *lu_fsm_state_tdef;
+
 struct lu_fsm_priv {
 	struct vlr_instance *vlr;
 	struct vlr_subscr *vsub;
@@ -675,7 +705,7 @@ struct lu_fsm_priv {
 	enum vlr_fsm_result result;
 	uint8_t rej_cause;
 
-	enum vlr_lu_type type;
+	enum vlr_lu_type lu_type;
 	bool lu_by_tmsi;
 	char imsi[16];
 	uint32_t tmsi;
@@ -762,7 +792,7 @@ static void lu_fsm_success(struct osmo_fsm_inst *fi)
 static void lu_fsm_failure(struct osmo_fsm_inst *fi, enum gsm48_reject_value rej_cause)
 {
 	struct lu_fsm_priv *lfp = lu_fsm_fi_priv(fi);
-	lfp->vlr->ops.tx_lu_rej(lfp->msc_conn_ref, rej_cause ? : GSM48_REJECT_NETWORK_FAILURE);
+	lfp->vlr->ops.tx_lu_rej(lfp->msc_conn_ref, rej_cause ? : GSM48_REJECT_NETWORK_FAILURE, lfp->lu_type);
 	_lu_fsm_done(fi, VLR_FSM_RESULT_FAILURE);
 }
 
@@ -773,7 +803,8 @@ static void vlr_loc_upd_start_lu_compl_fsm(struct osmo_fsm_inst *fi)
 		lu_compl_vlr_proc_alloc(fi, lfp->vsub, lfp->msc_conn_ref,
 					VLR_ULA_E_LU_COMPL_SUCCESS,
 					VLR_ULA_E_LU_COMPL_FAILURE,
-					lfp->assign_tmsi);
+					lfp->assign_tmsi,
+					lfp->lu_type);
 
 	osmo_fsm_inst_dispatch(lfp->lu_compl_vlr_fsm, LU_COMPL_VLR_E_START, NULL);
 }
@@ -939,7 +970,7 @@ static void vlr_loc_upd_node1_pre(struct osmo_fsm_inst *fi)
 	LOGPFSM(fi, "%s()\n", __func__);
 
 	if (vlr->cfg.check_imei_rqd && vlr->cfg.retrieve_imeisv_early) {
-		osmo_fsm_inst_state_chg(fi, VLR_ULA_S_WAIT_HLR_CHECK_IMEI_EARLY, vlr_timer(lfp->vlr, 3270), 3270);
+		osmo_tdef_fsm_inst_state_chg(fi, VLR_ULA_S_WAIT_HLR_CHECK_IMEI_EARLY, lu_fsm_state_tdef, vlr_tdefs, -1);
 		vlr_subscr_tx_req_check_imei(lfp->vsub);
 	} else {
 		vlr_loc_upd_node1(fi);
@@ -974,8 +1005,7 @@ static void vlr_loc_upd_want_imsi(struct osmo_fsm_inst *fi)
 	OSMO_ASSERT(lfp->vsub);
 
 	/* Obtain_IMSI_VLR */
-	osmo_fsm_inst_state_chg(fi, VLR_ULA_S_WAIT_IMSI,
-				vlr_timer(vlr, 3270), 3270);
+	osmo_tdef_fsm_inst_state_chg(fi, VLR_ULA_S_WAIT_IMSI, lu_fsm_state_tdef, vlr_tdefs, -1);
 	vlr->ops.tx_id_req(lfp->msc_conn_ref, GSM_MI_TYPE_IMSI);
 	/* will continue at vlr_loc_upd_node1_pre() once IMSI arrives */
 }
@@ -1106,13 +1136,12 @@ static void lu_fsm_idle(struct osmo_fsm_inst *fi, uint32_t event,
 
 	/* See 3GPP TS 23.012, procedure Retrieve_IMEISV_If_Required */
 	if ((!vlr->cfg.retrieve_imeisv_early)
-	    || (lfp->type == VLR_LU_TYPE_PERIODIC && lfp->vsub->imeisv[0])) {
+	    || (lfp->lu_type == VLR_LU_TYPE_PERIODIC && lfp->vsub->imeisv[0])) {
 		/* R_IMEISV_IR1 passed */
 		_start_lu_main(fi);
 	} else {
 		vlr->ops.tx_id_req(lfp->msc_conn_ref, GSM_MI_TYPE_IMEISV);
-		osmo_fsm_inst_state_chg(fi, VLR_ULA_S_WAIT_IMEISV,
-					vlr_timer(vlr, 3270), 3270);
+		osmo_tdef_fsm_inst_state_chg(fi, VLR_ULA_S_WAIT_IMEISV, lu_fsm_state_tdef, vlr_tdefs, -1);
 	}
 }
 
@@ -1307,7 +1336,7 @@ static void lu_fsm_wait_lu_compl(struct osmo_fsm_inst *fi, uint32_t event,
 		/* TODO: Notify_gsmSCF 23.078 */
 		/* TODO: Authenticated Radio Contact Established -> ARC */
 
-		if (lfp->type == VLR_LU_TYPE_IMSI_ATTACH)
+		if (lfp->lu_type == VLR_LU_TYPE_IMSI_ATTACH)
 			lfp->vlr->ops.tx_mm_info(lfp->msc_conn_ref);
 
 		lu_fsm_success(fi);
@@ -1488,7 +1517,7 @@ static struct osmo_fsm vlr_lu_fsm = {
 	.num_states = ARRAY_SIZE(vlr_lu_fsm_states),
 	.allstate_event_mask = 0,
 	.allstate_action = NULL,
-	.log_subsys = DVLR,
+	.log_subsys = DLGLOBAL,
 	.event_names = fsm_lu_event_names,
 	.cleanup = fsm_lu_cleanup,
 };
@@ -1529,7 +1558,7 @@ vlr_loc_update(struct osmo_fsm_inst *parent,
 	lfp->vlr = vlr;
 	lfp->msc_conn_ref = msc_conn_ref;
 	lfp->tmsi = tmsi;
-	lfp->type = type;
+	lfp->lu_type = type;
 	lfp->old_lai = *old_lai;
 	lfp->new_lai = *new_lai;
 	lfp->lu_by_tmsi = true;
@@ -1584,10 +1613,26 @@ void vlr_loc_update_cancel(struct osmo_fsm_inst *fi,
 		lu_fsm_failure(fi, gsm48_cause);
 }
 
-void vlr_lu_fsm_init(void)
+void vlr_lu_fsm_init(bool is_ps)
 {
+	if (is_ps) {
+		lu_fsm_state_tdef = sgsn_lu_tdef_states;
+		lu_compl_fsm_state_tdef = sgsn_lu_compl_tdef_states;
+	} else {
+		lu_fsm_state_tdef = msc_lu_tdef_states;
+		lu_compl_fsm_state_tdef = msc_lu_compl_tdef_states;
+	}
+
 	OSMO_ASSERT(osmo_fsm_register(&vlr_lu_fsm) == 0);
 	OSMO_ASSERT(osmo_fsm_register(&upd_hlr_vlr_fsm) == 0);
 	OSMO_ASSERT(osmo_fsm_register(&sub_pres_vlr_fsm) == 0);
 	OSMO_ASSERT(osmo_fsm_register(&lu_compl_vlr_fsm) == 0);
 }
+
+void vlr_lu_fsm_set_log_subsys(int log_subsys)
+{
+	vlr_lu_fsm.log_subsys = log_subsys;
+	upd_hlr_vlr_fsm.log_subsys = log_subsys;
+	sub_pres_vlr_fsm.log_subsys = log_subsys;
+	lu_compl_vlr_fsm.log_subsys = log_subsys;
+}
